/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "likelihoodcomponents.hpp"

#include "ublasutils/invert.hpp"
#include "ublasutils/determinant.hpp"


namespace myfitter {

using ublasutils::invert;
using boost::numeric::ublas::prod;
using boost::numeric::ublas::inner_prod;


// class GaussianLC

double GaussianLC::calc(const ObservableVector& pred,
                        const ObservableVector& meas)
{
    assert(int(pred.size()) >= nobservables());
    assert(int(meas.size()) >= nobservables());
        
    double d = pred[external_index_(0)]
        -meas[external_index_(0)];
    if(d < -syst_minus_)
        return std::pow((d + syst_minus_)/stat_err_, 2)
            + std::log(stat_err_*stat_err_*2.*M_PI);
    else if(d > syst_plus_)
        return std::pow((d - syst_plus_)/stat_err_, 2)
            + std::log(stat_err_*stat_err_*2.*M_PI);
    else
        return std::log(stat_err_*stat_err_*2.*M_PI);
}

int GaussianLC::calc_deriv(const ObservableVector& pred,
                           const ObservableVector& meas,
                           ObservableVector& result)
{
    assert(int(pred.size()) >= nobservables());
    assert(int(meas.size()) >= nobservables());
        
    double d = pred[external_index_(0)]
        -meas[external_index_(0)];
    if(d < -syst_minus_)
        result[external_index_(0)] = 2.*(d + syst_minus_)/(stat_err_*stat_err_);
    else if(d > syst_plus_)
        result[external_index_(0)] = 2.*(d - syst_plus_)/(stat_err_*stat_err_);
    else
        result[external_index_(0)] = 0.;
    return 0;
}

double GaussianLC::calc_nosystematics(const ObservableVector& pred,
                                      const ObservableVector& meas)
{
    assert(int(pred.size()) >= nobservables());
    assert(int(meas.size()) >= nobservables());
        
    double d = pred[external_index_(0)] - meas[external_index_(0)];
    return std::pow(d/stat_err_, 2) + std::log(stat_err_*stat_err_*2.*M_PI);
}


// class AsymmetricGaussianLC

double AsymmetricGaussianLC::calc(const ObservableVector& pred,
                                  const ObservableVector& meas)
{
    assert(int(pred.size()) >= nobservables());
    assert(int(meas.size()) >= nobservables());
        
    double stat_err = 0.5*(stat_plus_ + stat_minus_);

    double d = pred[external_index_(0)] - meas[external_index_(0)];
    if(d < -syst_minus_)
        return std::pow((d + syst_minus_)/stat_minus_, 2)
            + std::log(stat_err*stat_err*2.*M_PI);
    else if(d > syst_plus_)
        return std::pow((d - syst_plus_)/stat_plus_, 2)
            + std::log(stat_err*stat_err*2.*M_PI);
    else
        return std::log(stat_err*stat_err*2.*M_PI);
}

int AsymmetricGaussianLC::calc_deriv(const ObservableVector& pred,
                                     const ObservableVector& meas,
                                     ObservableVector& result)
{
    assert(int(pred.size()) >= nobservables());
    assert(int(meas.size()) >= nobservables());
        
    double d = pred[external_index_(0)]
        -meas[external_index_(0)];
    if(d < -syst_minus_)
        result[external_index_(0)]
            = 2.*(d + syst_minus_)/(stat_minus_*stat_minus_);
    else if(d > syst_plus_)
        result[external_index_(0)]
            = 2.*(d - syst_plus_)/(stat_plus_*stat_plus_);
    else
        result[external_index_(0)] = 0.;
    return 0;
}

double AsymmetricGaussianLC::calc_nosystematics(const ObservableVector& pred,
                                                const ObservableVector& meas)
{
    assert(int(pred.size()) >= nobservables());
    assert(int(meas.size()) >= nobservables());

    double stat_err = 0.5*(stat_plus_ + stat_minus_);
        
    double d = pred[external_index_(0)] - meas[external_index_(0)];
    if(d < 0.)
        return std::pow(d/stat_minus_, 2)
            + std::log(stat_err*stat_err*2.*M_PI);
    else
        return std::pow(d/stat_plus_, 2)
            + std::log(stat_err*stat_err*2.*M_PI);
}


// class CorrelatedGaussianLC

void CorrelatedGaussianLC::add(int iobs, double val, double err)
{
    add_(iobs, val);
    int n = n_claimed_observables();
    err_.resize(n);
    err_[n-1] = err;
    cor_.resize(n, n);
    for(int i=0; i<n-2; ++i)
        cor_(i, n-1) = cor_(n-1, i) = 0.;
    cor_(n-1, n-1) = 1.;
    invvar_.resize(n, n);
    needs_update_ = true;
}

void CorrelatedGaussianLC::update_()
{
    int n = n_claimed_observables();
    Matrix var(n, n);
    for(int i=0; i<n; ++i)
        for(int j=i; j<n; ++j)
            var(i,j) = var(j,i) = cor_(i,j)*err_[i]*err_[j];

    invert(var, invvar_);
    needs_update_ = false;
}

double CorrelatedGaussianLC::calc(const ObservableVector& pred,
                                  const ObservableVector& meas)
{
    if(needs_update_) update_();
    int n = n_claimed_observables();
    boost::numeric::ublas::vector<double> d(n);
    for(int i=0; i<n; ++i)
        d(i) = pred[external_index_(i)] - meas[external_index_(i)];
    boost::numeric::ublas::vector<double> invvd(prod(invvar_, d));
    double detinvvar = myfitter::ublasutils::det(invvar_);
    return inner_prod(d, invvd) + std::log(std::pow(2.*M_PI, n)/detinvvar);
}

int CorrelatedGaussianLC::calc_deriv(const ObservableVector& pred,
                                     const ObservableVector& meas,
                                     ObservableVector& result)
{
    if(needs_update_) update_();
    int n = n_claimed_observables();
    boost::numeric::ublas::vector<double> d(n);
    for(int i=0; i<n; ++i)
        d(i) = pred[external_index_(i)] - meas[external_index_(i)];
    boost::numeric::ublas::vector<double> invvd(prod(invvar_, d));
    for(int i=0; i<n; ++i)
        result[external_index_(i)] = 2.*invvd[i];
    return 0;
}

void CorrelatedGaussianLC::get_hessian(Matrix& hess,
                                       const ObservableVector& obs)
{
    if(needs_update_) update_();
    int n = n_claimed_observables();
    for(int i=0; i<n; ++i)
        for(int j=0; j<n; ++j)
            hess(external_index_(i),external_index_(j))
                = invvar_(i,j);
}


}
