#include <iostream>

#include "fitter.hpp"
#include "likelihoodcomponents.hpp"
#include "comparison.hpp"

#include <boost/numeric/ublas/io.hpp>

using namespace myfitter;

class LinearModel : public Model {
public:
    enum {P1, P2, P3, P4, NPAR};
    enum {O1, O2, O3, O4, O5, O6, O7, NOBS};

    Matrix mat;
    ObservableVector vec;

    LinearModel() : Model(NPAR,NOBS), mat(NOBS,NPAR), vec(NOBS) { }
    LinearModel(const LinearModel& m) : Model(m), mat(m.mat), vec(m.vec) { }
    virtual const LinearModel& operator=(const LinearModel& m) {
        Model::operator=(m);
        mat = m.mat;
        vec = m.vec;
        return *this;
    }

    virtual int calc() {
        observables_
            .assign(boost::numeric::ublas::prod(mat, parameters()) + vec);
        needs_update_ = false;
        return 0;
    }
};

int main()
{
    typedef LinearModel Mod;


    // initialize model

    LinearModel model;

    model.derivative_epsilon(1.e-7);

    model.scale(Mod::P1, 0.1);
    model.scale(Mod::P2, 0.1);
    model.scale(Mod::P3, 0.1);
    model.scale(Mod::P4, 0.1);

    model.scan_range(Mod::P1, -2.5, 2.5);
    model.scan_range(Mod::P2, -2.5, 2.5);
    model.scan_range(Mod::P3, -2.5, 2.5);
    model.scan_range(Mod::P4, -2.5, 2.5);

    model.mat(Mod::O1, Mod::P1) = +0.3;
    model.mat(Mod::O1, Mod::P2) = -1.2;
    model.mat(Mod::O1, Mod::P3) = +0.0;
    model.mat(Mod::O1, Mod::P4) = -0.7;

    model.mat(Mod::O2, Mod::P1) = +0.4;
    model.mat(Mod::O2, Mod::P2) = +0.6;
    model.mat(Mod::O2, Mod::P3) = -0.9;
    model.mat(Mod::O2, Mod::P4) = +1.0;

    model.mat(Mod::O3, Mod::P1) = -1.5;
    model.mat(Mod::O3, Mod::P2) = -0.1;
    model.mat(Mod::O3, Mod::P3) = -0.1;
    model.mat(Mod::O3, Mod::P4) = -0.5;

    model.mat(Mod::O4, Mod::P1) = +0.0;
    model.mat(Mod::O4, Mod::P2) = +0.0;
    model.mat(Mod::O4, Mod::P3) = -0.6;
    model.mat(Mod::O4, Mod::P4) = +0.2;

    model.mat(Mod::O5, Mod::P1) = +0.4;
    model.mat(Mod::O5, Mod::P2) = -0.4;
    model.mat(Mod::O5, Mod::P3) = -1.1;
    model.mat(Mod::O5, Mod::P4) = +0.7;

    model.mat(Mod::O6, Mod::P1) = -0.3;
    model.mat(Mod::O6, Mod::P2) = -0.6;
    model.mat(Mod::O6, Mod::P3) = +0.2;
    model.mat(Mod::O6, Mod::P4) = +0.9;

    model.mat(Mod::O7, Mod::P1) = -0.1;
    model.mat(Mod::O7, Mod::P2) = +0.9;
    model.mat(Mod::O7, Mod::P3) = -1.0;
    model.mat(Mod::O7, Mod::P4) = +0.0;

    model.vec(Mod::O1) = +0.5;
    model.vec(Mod::O2) = +0.5;
    model.vec(Mod::O3) = -0.4;
    model.vec(Mod::O4) = -0.8;
    model.vec(Mod::O5) = -0.9;
    model.vec(Mod::O6) = +0.2;
    model.vec(Mod::O7) = +0.0;


    // construct inputs

    ParameterVector truepar(Mod::NPAR);
    truepar(Mod::P1) = -1.0;
    truepar(Mod::P2) = -2.0;
    truepar(Mod::P3) = +0.4;
    truepar(Mod::P4) = +0.9;
    model.parameters(truepar);

    ObservableVector trueobs(Mod::NOBS);
    model.calc();
    trueobs = model.observables();


    // initialise fitter and likelihood function

    Fitter fitter(Mod::NOBS);
    fitter.dvegas_precision(0.02);
    fitter.dvegas_nfirstshots(2000);
    fitter.dvegas_nshots(1000);

    fitter.likelihood_function()
        .add(GaussianLC(Mod::O1, trueobs(Mod::O1)+0.3, 0.2));
    fitter.likelihood_function()
        .add(GaussianLC(Mod::O2, trueobs(Mod::O2)-0.1, 0.2));
    fitter.likelihood_function()
        .add(GaussianLC(Mod::O4, trueobs(Mod::O4)-0.5, 0.3));
    fitter.likelihood_function()
        .add(GaussianLC(Mod::O6, trueobs(Mod::O6)+0.2, 0.1));
    
    CorrelatedGaussianLC lc(Mod::NOBS);
    lc.add(Mod::O3, trueobs(Mod::O3)-0.5, 0.2);
    lc.add(Mod::O5, trueobs(Mod::O5)+0.1, 0.3);
    lc.add(Mod::O7, trueobs(Mod::O7)-0.2, 0.2);
    lc.cor(Mod::O3, Mod::O5, -0.1);
    lc.cor(Mod::O3, Mod::O7, -0.5);
    lc.cor(Mod::O5, Mod::O7, +0.3);
    fitter.likelihood_function().add(lc);

    ObservableVector cvals(Mod::NOBS);
    fitter.likelihood_function().get_central_values(cvals);

    
    // scan full and constrained model

    model.clear();
    model.scan(1000);
    if(fitter.global_fit(model)) return 1;
    if(differ(model.chisquare(), 2.87934, 1.e-5, 1.e-10)) return 1;
    if(differ(model.parameter(Mod::P1), -0.770328, 1.e-5, 1.e-10)) return 1;
    if(differ(model.parameter(Mod::P2), -2.1953, 1.e-5, 1.e-10)) return 1;
    if(differ(model.parameter(Mod::P3), 0.573968, 1.e-5, 1.e-10)) return 1;
    if(differ(model.parameter(Mod::P4), 1.00008, 1.e-5, 1.e-10)) return 1;

    LinearModel cmod = model;
    cmod.parameter(Mod::P1, -0.7); cmod.fix(Mod::P1);
    cmod.parameter(Mod::P2, -2.0); cmod.fix(Mod::P2);
    cmod.clear();
    cmod.scan(100);
    if(fitter.global_fit(cmod)) return 1;
    if(differ(cmod.chisquare(), 6.11058, 1.e-5, 1.e-10)) return 1;
    if(differ(cmod.parameter(Mod::P3), 0.697862, 1.e-5, 1.e-10)) return 1;
    if(differ(cmod.parameter(Mod::P4), 1.04248, 1.e-5, 1.e-10)) return 1;


    // calculate relative p-value

    fitter.calc_lrt_pvalue(model, cmod);
    if(!fitter.reached_precision_goal()) return 1;
    if(fitter.failed_shot_ratio() > 0.005) return 1;
    if(differ(fitter.pvalue(), 0.198767, 0.05, 1.e-10)) return 1;

    return 0;
}
